@@ -3,7 +3,7 @@
 # @Date     : 2020/06/01
 # @Author   : mingming.xu
 # @Email    : xv44586@gmail.com
-
+import numpy as np
 import tensorflow as tf
 
 from toolkit4nlp.backend import K, keras
@@ -26,7 +26,7 @@ class MultiHeadAttention(Layer):
     """
 
     def __init__(self, head_nums, head_size, key_size=None, use_bias=True, attention_scale=True,
-                 kernel_initializer='glorot_uniform', **kwargs):
+                 kernel_initializer='glorot_uniform', with_residual_attention=False, **kwargs):
         super(MultiHeadAttention, self).__init__(**kwargs)
         self.head_nums = head_nums
         self.head_size = head_size
@@ -35,6 +35,7 @@ class MultiHeadAttention(Layer):
         self.use_bias = use_bias
         self.attention_scale = attention_scale
         self.kernel_initializer = initializers.get(kernel_initializer)
+        self.with_residual_attention = with_residual_attention  # realformer
 
     def build(self, input_shape):
         super(MultiHeadAttention, self).build(input_shape)
@@ -60,14 +61,14 @@ class MultiHeadAttention(Layer):
             kernel_initializer=self.kernel_initializer
         )
 
-    def call(self, inputs, mask=None, a_mask=None, position_bias=None):
+    def call(self, inputs, mask=None, **kwargs):
         """
         多头注意力
-        :param inputs: [q, k, v, a_mask, position_bias]
-        :param mask: [q_mask, v_mask],
+        :param inputs: [q, k, v, a_bias, position_bias]
+        :param mask: [q_mask, k_mask, v_mask],
             q_mask 对query序列进行mask，针对padding；v_mask对value序列进行mask，防止看到某些位置value，如padding
-        :param a_mask: Boolean，是否对attention进行mask
-        :param position_bias: type of position bias， 使用指定类型的位置编码对attention里的位置进行偏移
+        :param a_bias: Boolean，是否对attention进行偏移，包括attention mask/ pre attention scores
+        :param position_bias: type of position bias， 使用指定类型的位置编码对attention里的位置信息进行偏移
         :return:
         """
         q, k, v = inputs[:3]
@@ -77,8 +78,9 @@ class MultiHeadAttention(Layer):
                 q_mask = K.cast(mask[0], K.floatx())
             if mask[2] is not None:
                 v_mask = K.cast(mask[2], K.floatx())
-        if a_mask is not None:
-            a_mask = inputs[idx]
+        attention_bias, position_bias = kwargs.get('a_bias'), kwargs.get('position_bias')
+        if attention_bias:
+            attention_bias = inputs[idx]
             idx += 1
 
         # 投影变换
@@ -102,9 +104,11 @@ class MultiHeadAttention(Layer):
 
         # value mask
         att = sequence_masking(att, v_mask, 'add', -1)
-        # attention mask
-        if a_mask is not None:
-            att = att - (1 - a_mask) * 1e12
+
+        # attention mask / residual attention scores
+        if attention_bias is not None:
+            att += attention_bias
+        attention_bias = att
 
         att = K.softmax(att)
         output = tf.einsum('bhjk,bkhd->bjhd', att, vw)
@@ -113,15 +117,26 @@ class MultiHeadAttention(Layer):
             output = output + tf.einsum('bhjk,jkd->bjhd', att, position_embeddings)
         output = K.reshape(output, (-1, K.shape(output)[1], self.output_dim))
         output = self.combine_dense(output)
+
         # query mask
-        output = sequence_masking(output, q_mask, 'mul')
-        return output
+        # output = sequence_masking(output, q_mask, 'mul')
+        if not self.with_residual_attention:
+            return output
+
+        return [output, attention_bias]
 
     def compute_output_shape(self, input_shape):
-        return (input_shape[0][0], input_shape[0][1], self.output_dim)
+        output_shape = (input_shape[0][0], input_shape[0][1], self.output_dim)
+        if self.with_residual_attention:
+            att_shape = (input_shape[0][0], self.head_nums, input_shape[0][1], input_shape[1][1])
+            return [output_shape, att_shape]
+
+        return output_shape
 
     def compute_mask(self, inputs, mask=None):
         if mask is not None:
+            if self.with_residual_attention:
+                return [mask[0], None]
             return mask[0]
 
     def get_config(self):
@@ -131,6 +146,7 @@ class MultiHeadAttention(Layer):
                        'key_size': self.key_size,
                        'use_bias': self.use_bias,
                        'attention_scale': self.attention_scale,
+                       'with_residual_attention': self.with_residual_attention,
                        'kernel_initializer': initializers.serialize(self.kernel_initializer)})
         return config
 
@@ -142,21 +158,69 @@ class LayerNormalization(Layer):
     std = sqrt(mean(square(x - x_mean)) + epsilon)
     x = beta + (x - x_mean / std) * gamma
     """
+    """
+    增加Conditional，用来控制模型的表现。主要思路为通过调整center 与scale，来调整模型的输出，控制其表现。
+    condition_* 为Conditional 系列参数，当Conditional=True时生效。
+    ref: [基于Conditional Layer Normalization的条件文本生成](https://kexue.fm/archives/7124)
+    """
 
-    def __init__(self, center=True, scale=True, epsilon=None, **kwargs):
+    def __init__(self, center=True, scale=True, epsilon=None,
+                 conditional=False, condition_hidden_units=None, condition_hidden_activation='linear',
+                 condition_hidden_initializer='glorot_uniform', **kwargs):
         super(LayerNormalization, self).__init__(**kwargs)
         self.center = center
         self.scale = scale
         self.epsilon = epsilon or 1e-12
+        self.conditional = conditional
+        self.condition_hidden_units = condition_hidden_units
+        self.condition_hidden_activation = activations.get(condition_hidden_activation)
+        self.condition_hidden_initializer = initializers.get(condition_hidden_initializer)
 
     def build(self, input_shape):
         super(LayerNormalization, self).build(input_shape)
+
+        if self.conditional:
+            shape = (input_shape[0][-1], )
+        else:
+            shape = (input_shape[-1], )
+
         if self.center:
-            self.beta = self.add_weight(shape=(input_shape[-1],), initializer='zeros', name='beta')
+            self.beta = self.add_weight(shape=shape, initializer='zeros', name='beta')
         if self.scale:
-            self.gamma = self.add_weight(shape=(input_shape[-1],), initializer='ones', name='gamma')
+            self.gamma = self.add_weight(shape=shape, initializer='ones', name='gamma')
+
+        if self.conditional:
+            # hidden units 不为None，做一次维度转换
+            if self.condition_hidden_units is not None:
+                self.condition_hidden_dense = Dense(units=self.condition_hidden_units,
+                                                    activation=self.condition_hidden_activation,
+                                                    use_bias=False,
+                                                    kernel_initializer=self.condition_hidden_initializer)
+            # 用0初始化center_dense 与scale_dense ,防止扰乱原模型
+            if self.center:
+                self.beta_dense = Dense(units=shape[0], use_bias=False, kernel_initializer='zeros')
+            if self.scale:
+                self.gamma_dense = Dense(units=shape[0], use_bias=False, kernel_initializer='zeros')
 
     def call(self, inputs):
+        """
+        conditional 时， condition 放在inputs后面，[inputs, condition]
+        """
+        if self.conditional:
+            inputs, cond = inputs
+            if self.condition_hidden_units is not None:
+                cond = self.condition_hidden_dense(cond)
+            # 适配cond维度，与inputs保持一致
+            for _ in range(K.ndim(inputs) - K.ndim(cond)):
+                cond = K.expand_dims(cond, 1)
+            if self.center:
+                beta = self.beta_dense(cond) + self.beta
+            if self.scale:
+                gamma = self.gamma_dense(cond) + self.gamma
+        else:
+            beta = self.beta
+            gamma = self.gamma
+
         output = inputs
         if self.center:
             mean = K.mean(inputs, axis=-1, keepdims=True)
@@ -166,21 +230,39 @@ class LayerNormalization(Layer):
             var = K.mean(K.square(output), axis=-1, keepdims=True)
             std = K.sqrt(var + self.epsilon)
             output = output / std
-            output = output * self.gamma
+            output = output * gamma
 
         if self.center:
-            output = output + self.beta
+            output = output + beta
 
         return output
 
     def compute_mask(self, inputs, mask=None):
+        if self.conditional:
+            masks = mask if mask is not None else []
+            masks = [K.expand_dims(m, 0) for m in masks if m is not None]
+            if len(masks) == 0:
+                return None
+            else:
+                return K.all(K.concatenate(masks, axis=0), axis=0)
+
         return mask
 
+    def compute_output_shape(self, input_shape):
+        if self.conditional:
+            return input_shape[0]
+        else:
+            return input_shape
+
     def get_config(self):
         base_config = super(LayerNormalization, self).get_config()
         base_config.update({"center": self.center,
                             "scale": self.scale,
-                            "epsilon": self.epsilon})
+                            "epsilon": self.epsilon,
+                            "conditional": self.conditional,
+                            "condition_hidden_units": self.condition_hidden_units,
+                            "condition_hidden_activation": activations.serialize(self.condition_hidden_activation),
+                            "condition_hidden_initializer": initializers.serialize(self.condition_hidden_initializer)})
         return base_config
 
 
@@ -444,6 +526,127 @@ class RelativePositionEmbedding(Layer):
         return dict(list(base_config.items()) + list(config.items()))
 
 
+class RelativePositionEmbeddingT5(RelativePositionEmbedding):
+    """Google T5的相对位置编码
+    来自论文：https://arxiv.org/abs/1910.10683
+    """
+    def __init__(
+        self,
+        input_dim,
+        output_dim,
+        max_distance=128,
+        bidirectional=True,
+        embeddings_initializer='zeros',
+        **kwargs
+    ):
+        super(RelativePositionEmbeddingT5,
+              self).__init__(input_dim, output_dim, **kwargs)
+        self.max_distance = max_distance
+        self.bidirectional = bidirectional
+
+    def compute_position_ids(self, inputs):
+        """T5的相对位置分桶（直接翻译自官方T5源码）
+        i-i:   0 1 2 3 4 5 6 7 8 9 10 11 12 13 14...
+        f(i-j):0 1 2 3 4 5 6 7 8 8 8  8  9   9  9 ...
+        """
+        q, v = inputs
+        # 计算位置差
+        q_idxs = K.arange(0, K.shape(q)[1], dtype='int32')
+        q_idxs = K.expand_dims(q_idxs, 1)
+        v_idxs = K.arange(0, K.shape(v)[1], dtype='int32')
+        v_idxs = K.expand_dims(v_idxs, 0)
+        pos_ids = v_idxs - q_idxs
+        # 后处理操作
+        num_buckets, max_distance = self.input_dim, self.max_distance
+        ret = 0
+        n = -pos_ids
+        if self.bidirectional:
+            num_buckets //= 2
+            ret += K.cast(K.less(n, 0), 'int32') * num_buckets
+            n = K.abs(n)
+        else:
+            n = K.maximum(n, 0)
+        # now n is in the range [0, inf)
+        max_exact = num_buckets // 2
+        is_small = K.less(n, max_exact)
+        val_if_large = max_exact + K.cast(
+            K.log(K.cast(n, K.floatx()) / max_exact) /
+            np.log(max_distance / max_exact) * (num_buckets - max_exact),
+            'int32',
+        )
+        val_if_large = K.minimum(val_if_large, num_buckets - 1)
+        ret += K.switch(is_small, n, val_if_large)
+        return ret
+
+    def get_config(self):
+        config = {
+            'max_distance': self.max_distance,
+            'bidirectional': self.bidirectional,
+        }
+        base_config = super(RelativePositionEmbeddingT5, self).get_config()
+        return dict(list(base_config.items()) + list(config.items()))
+
+
+class AbsolutePositionEmbeddingTUPE(Layer):
+    """
+    TUPE式位置与位置交互的位置编码
+    出自论文：http://arxiv.org/abs/2006.15595
+    """
+    def __init__(self,
+                 input_dim,
+                 output_dim,
+                 num_attention_heads,
+                 attn_scale_factor=2,
+                 relative_position_bias=True,
+                 embeddings_initializer='zeros',
+                 **kwargs
+                 ):
+        super(AbsolutePositionEmbeddingTUPE, self).__init__(**kwargs)
+        self.input_dim = input_dim
+        self.output_dim = output_dim
+        self.num_attention_heads = num_attention_heads
+        self.attn_scale_factor = attn_scale_factor
+        self.relative_position_bias = relative_position_bias
+        self.embedding_initializer = initializers.get(embeddings_initializer)
+        self.pos_scaling = float(self.output_dim / num_attention_heads * self.attn_scale_factor) ** -0.5
+
+    def build(self, input_shape):
+        self.embeddings = self.add_weight(name='absolutePositionEmbedding',
+                                          shape=(self.input_dim, self.output_dim),
+                                          initializer=self.embedding_initializer, )
+        self.q_dense = Dense(self.output_dim, use_bias=False)
+        self.k_dense = Dense(self.output_dim, use_bias=False)
+        self.pos_ln = LayerNormalization()
+
+    def call(self, inputs, **kwargs):
+        q, v = inputs
+        seq_len = K.shape(q)[1]
+        position_ids = K.arange(0, seq_len + 1, 'int32')  # 增加一个虚拟头结点来解耦cls
+        x = K.gather(self.embeddings, position_ids)
+        x = self.pos_ln(x)
+        q = self.q_dense(x) * self.pos_scaling
+        k = self.k_dense(x)
+        q = K.reshape(q, (seq_len+1, self.num_attention_heads, -1))
+        k = K.reshape(k, (seq_len+1, self.num_attention_heads, -1))
+
+        abs_pos_bias = tf.einsum('jhd,khd->hjk', q, k)
+        # p_0 \dot p_0 is  cla to others
+        cls_2_other = abs_pos_bias[:, 0, 0]
+        # p_1 \dot p_1 is others to cls
+        other_2_cls = abs_pos_bias[:, 1, 1]
+        # offset
+        abs_pos_bias = abs_pos_bias[:, 1:, 1:]
+        abs_pos_bias[:, :, 0] = K.reshape(other_2_cls, (-1, 1))
+        abs_pos_bias[:, 0, :] = K.reshape(cls_2_other, (-1, 1))
+        if self.relative_position_bias:
+            rel_pos_bias = self.compute_rel_pos_bias(inputs)
+            abs_pos_bias += rel_pos_bias
+        return abs_pos_bias
+
+    def compute_rel_pos_bias(self, inputs):
+        pass
+
+
 class DGCNN(Layer):
     """
     膨胀卷积网络，优势是本质是一个CNN，所以速度上比RNNs快，同时通过不同的膨胀系数，如【1，3，5，8】可以来整合全局信息，
@@ -672,6 +875,47 @@ class ConditionalRandomField(Layer):
         return dict(list(base_config.items()) + list(config.items()))
 
 
+class GatherLastToken(Layer):
+    """
+    获取最后一个token对应的output，
+    如果指定pad_token_id，则是最后一个非pad_token对应的output，否则为序列最后一个token对应的output
+    """
+
+    def __init__(self, pad_token_id=None, **kwargs):
+        super(GatherLastToken, self).__init__(**kwargs)
+        self.pad_token_id = pad_token_id
+
+    def call(self, inputs, **kwargs):
+        logits, token_seq = inputs[:2]
+        seq_shape = K.shape(token_seq)
+        batch_size, seq_length = seq_shape[0], seq_shape[1]
+        if self.pad_token_id is None:
+            sequence_lengths = [seq_length - 1] * batch_size
+        else:
+            sequence_lengths = (
+                    K.sum(
+                        K.cast(
+                            K.not_equal(token_seq, self.pad_token_id),
+                            dtype='int32',
+                        ),
+                        -1,
+                        keepdims=False,
+                    )
+                    - 1
+            )
+        # only tf2
+        # return tf.gather(logits, sequence_lengths, batch_dims=1, axis=1)
+
+        indices = K.expand_dims(sequence_lengths, -1)
+        return tf.gather_nd(logits, indices, batch_dims=1)
+
+    def compute_output_shape(self, input_shape):
+        return (input_shape[0][0], input_shape[0][-1])
+
+    def compute_mask(self, inputs, mask=None):
+        return mask[0]
+
+
 class Loss(Layer):
     """
     特殊层，用于编写灵活loss
@@ -708,6 +952,20 @@ class Loss(Layer):
         return dict(list(base_config.items()) + list(config.items()))
 
 
+class ReWeight(Layer):
+    def __init__(self, init_reweight=0, trainable=True, **kwargs):
+        super(ReWeight, self).__init__(**kwargs)
+        self.init_reweight = init_reweight
+        self.trainable = trainable
+
+    def build(self, input_shape):
+        initializer = keras.initializers.Constant(self.init_reweight)
+        self.beta = self.add_weight(shape=(1,), initializer=initializer, name='reweight', trainable=self.trainable)
+
+    def call(self, x):
+        return x * self.beta
+
+
 custom_objects = {
     'Embedding': Embedding,
     'BiasAdd': BiasAdd,
@@ -721,7 +979,8 @@ custom_objects = {
     'DGCNN': DGCNN,
     'SinCosPositionEmbedding': SinCosPositionEmbedding,
     'Loss': Loss,
-    'RelativePositionEmbedding': RelativePositionEmbedding
+    'RelativePositionEmbedding': RelativePositionEmbedding,
+    'ReWeight': ReWeight,
 }
 
 keras.utils.get_custom_objects().update(custom_objects)
